@@ -29,6 +29,7 @@
 #include "tls.h"
 #include "checksum.h"
 #include "cipher.h"
+#include "cert.h"
 #include "tls-private.h"
 #include "random.h"
 
@@ -75,6 +76,7 @@ static void tls_tx_record_plaintext(struct l_tls *tls,
 	uint8_t buf[TX_RECORD_HEADROOM + TX_RECORD_MAX_LEN +
 				TX_RECORD_TAILROOM];
 	uint8_t iv[32];
+	uint8_t *assocdata;
 	int offset;
 
 	/*
@@ -133,14 +135,14 @@ static void tls_tx_record_plaintext(struct l_tls *tls,
 
 		offset = 0;
 
-		if (tls->negotiated_version >= TLS_V12) {
+		if (tls->negotiated_version >= L_TLS_V12) {
 			l_getrandom(ciphertext, tls->record_iv_length[1]);
 
 			l_cipher_set_iv(tls->cipher[1], ciphertext,
 					tls->record_iv_length[1]);
 
 			offset = tls->record_iv_length[1];
-		} else if (tls->negotiated_version >= TLS_V11) {
+		} else if (tls->negotiated_version >= L_TLS_V11) {
 			l_getrandom(iv, tls->record_iv_length[1]);
 
 			l_cipher_encrypt(tls->cipher[1], iv, ciphertext,
@@ -156,7 +158,49 @@ static void tls_tx_record_plaintext(struct l_tls *tls,
 		break;
 
 	case TLS_CIPHER_AEAD:
-		/* No AEAD ciphers supported today */
+		/* Prepend seq_num to TLSCompressed.type + .version + .length */
+		assocdata = compressed - 8;
+		l_put_be64(tls->seq_num[1]++, assocdata);
+
+		cipher_input = compressed + 5;
+		cipher_input_len = compressed_len;
+
+		/*
+		 * Build the IV.  The explicit part generation method is
+		 * actually cipher suite-specific but our only AEAD cipher
+		 * suites only require this part to be unique for each
+		 * record.  For future suites there may need to be a callback
+		 * that generates the per-record IV or an enum for the suite
+		 * to select one of a few IV types.
+		 *
+		 * Note kernel's rfc4106(gcm(...)) algorithm could potentially
+		 * be used to build the IV.
+		 */
+		memcpy(iv, tls->fixed_iv[1], tls->fixed_iv_length[1]);
+		l_put_le64(tls->seq_num[1], iv + tls->fixed_iv_length[1]);
+
+		if (tls->record_iv_length[1] > 8)
+			memset(iv + tls->fixed_iv_length[1] + 8, 42,
+				tls->record_iv_length[1] - 8);
+
+		/* Build the GenericAEADCipher struct */
+		ciphertext = buf + TX_RECORD_HEADROOM;
+		memcpy(ciphertext, iv + tls->fixed_iv_length[1],
+			tls->record_iv_length[1]);
+		l_aead_cipher_encrypt(tls->aead_cipher[1],
+					cipher_input, cipher_input_len,
+					assocdata, 13,
+					iv, tls->fixed_iv_length[1] +
+					tls->record_iv_length[1],
+					ciphertext + tls->record_iv_length[1],
+					cipher_input_len +
+					tls->auth_tag_length[1]);
+
+		ciphertext_len = tls->record_iv_length[1] +
+			cipher_input_len + tls->auth_tag_length[1];
+
+		break;
+
 	default:
 		return;
 	}
@@ -179,7 +223,7 @@ void tls_tx_record(struct l_tls *tls, enum tls_content_type type,
 				TX_RECORD_TAILROOM];
 	uint8_t *fragment, *plaintext;
 	uint16_t fragment_len;
-	uint16_t version = tls->negotiated_version ?: TLS_MIN_VERSION;
+	uint16_t version = tls->negotiated_version ?: tls->min_version;
 
 	if (type == TLS_CT_ALERT)
 		tls->record_flush = true;
@@ -290,6 +334,9 @@ static bool tls_handle_plaintext(struct l_tls *tls, const uint8_t *plaintext,
 
 				tls->message_buf_len = 0;
 
+				if (tls->record_flush)
+					break;
+
 				continue;
 			}
 
@@ -332,6 +379,7 @@ static bool tls_handle_ciphertext(struct l_tls *tls)
 	uint8_t *compressed;
 	int compressed_len;
 	uint8_t iv[32];
+	uint8_t *assocdata;
 
 	type = tls->record_buf[0];
 	version = l_get_be16(tls->record_buf + 1);
@@ -396,7 +444,7 @@ static bool tls_handle_ciphertext(struct l_tls *tls)
 
 	case TLS_CIPHER_BLOCK:
 		i = 0;
-		if (tls->negotiated_version >= TLS_V11)
+		if (tls->negotiated_version >= L_TLS_V11)
 			i = tls->record_iv_length[0];
 
 		if (fragment_len <= tls->mac_length[0] + i) {
@@ -422,7 +470,7 @@ static bool tls_handle_ciphertext(struct l_tls *tls)
 			return false;
 		}
 
-		if (tls->negotiated_version >= TLS_V12) {
+		if (tls->negotiated_version >= L_TLS_V12) {
 			if (!l_cipher_set_iv(tls->cipher[0],
 						tls->record_buf + 5,
 						tls->record_iv_length[0])) {
@@ -430,7 +478,7 @@ static bool tls_handle_ciphertext(struct l_tls *tls)
 						"Setting fragment IV failed");
 				return false;
 			}
-		} else if (tls->negotiated_version >= TLS_V11)
+		} else if (tls->negotiated_version >= L_TLS_V11)
 			if (!l_cipher_decrypt(tls->cipher[0],
 						tls->record_buf + 5, iv,
 						tls->record_iv_length[0])) {
@@ -474,10 +522,9 @@ static bool tls_handle_ciphertext(struct l_tls *tls)
 			tls->mac_length[0];
 		l_put_be16(compressed_len, compressed + 11);
 
-		for (i = 0; i < padding_len; i++)
-			if (compressed[13 + cipher_output_len - 1 -
-					padding_len + i] != padding_len)
-				error = 1;
+		error |= !l_secure_memeq(compressed + 13 + cipher_output_len -
+						1 - padding_len, padding_len,
+						padding_len);
 
 		/* Calculate the MAC if needed */
 		tls_write_mac(tls, compressed + 8, 5 + compressed_len,
@@ -496,7 +543,41 @@ static bool tls_handle_ciphertext(struct l_tls *tls)
 		break;
 
 	case TLS_CIPHER_AEAD:
-		/* No AEAD ciphers supported today */
+		if (fragment_len <= tls->record_iv_length[0] +
+				tls->auth_tag_length[0]) {
+			TLS_DISCONNECT(TLS_ALERT_DECODE_ERROR, 0,
+					"Record fragment too short: %u",
+					fragment_len);
+			return false;
+		}
+
+		compressed_len = fragment_len - tls->record_iv_length[0] -
+			tls->auth_tag_length[0];
+		l_put_be16(compressed_len, compressed + 11);
+
+		/* Prepend seq_num to TLSCompressed.type + .version + .length */
+		assocdata = compressed;
+		l_put_be64(tls->seq_num[0]++, assocdata);
+		compressed += 13;
+
+		/* Build the IV */
+		memcpy(iv, tls->fixed_iv[0], tls->fixed_iv_length[0]);
+		memcpy(iv + tls->fixed_iv_length[0], tls->record_buf + 5,
+			tls->record_iv_length[0]);
+
+		if (!l_aead_cipher_decrypt(tls->aead_cipher[0],
+				tls->record_buf + 5 + tls->record_iv_length[0],
+				fragment_len - tls->record_iv_length[0],
+				assocdata, 13, iv, tls->fixed_iv_length[0] +
+				tls->record_iv_length[0],
+				compressed, compressed_len)) {
+			TLS_DISCONNECT(TLS_ALERT_INTERNAL_ERROR, 0,
+					"Decrypting record fragment failed");
+			return false;
+		}
+
+		break;
+
 	default:
 		return false;
 	}
@@ -530,10 +611,8 @@ LIB_EXPORT void l_tls_handle_rx(struct l_tls *tls, const uint8_t *data,
 				tls->record_buf_len = 0;
 				need_len = 5;
 
-				if (tls->record_flush) {
-					tls->record_flush = false;
+				if (tls->record_flush)
 					break;
-				}
 			}
 
 			if (!len)
