@@ -26,11 +26,10 @@
 
 #define _GNU_SOURCE
 #include <stdio.h>
-#include <ctype.h>
-#include <stdarg.h>
 #include <stdlib.h>
 #include <limits.h>
 
+#include "utf8.h"
 #include "util.h"
 #include "private.h"
 
@@ -64,7 +63,7 @@ LIB_EXPORT void *l_malloc(size_t size)
 			return ptr;
 
 		fprintf(stderr, "%s:%s(): failed to allocate %zd bytes\n",
-					STRLOC, __PRETTY_FUNCTION__, size);
+					STRLOC, __func__, size);
 		abort();
 	}
 
@@ -94,7 +93,7 @@ LIB_EXPORT void *l_realloc(void *mem, size_t size)
 			return ptr;
 
 		fprintf(stderr, "%s:%s(): failed to re-allocate %zd bytes\n",
-					STRLOC, __PRETTY_FUNCTION__, size);
+					STRLOC, __func__, size);
 		abort();
 	} else
 		l_free(mem);
@@ -154,7 +153,7 @@ LIB_EXPORT char *l_strdup(const char *str)
 			return tmp;
 
 		fprintf(stderr, "%s:%s(): failed to allocate string\n",
-						STRLOC, __PRETTY_FUNCTION__);
+						STRLOC, __func__);
 		abort();
 	}
 
@@ -182,7 +181,7 @@ LIB_EXPORT char *l_strndup(const char *str, size_t max)
 			return tmp;
 
 		fprintf(stderr, "%s:%s(): failed to allocate string\n",
-						STRLOC, __PRETTY_FUNCTION__);
+						STRLOC, __func__);
 		abort();
 	}
 
@@ -208,7 +207,7 @@ LIB_EXPORT char *l_strdup_printf(const char *format, ...)
 
 	if (len < 0) {
 		fprintf(stderr, "%s:%s(): failed to allocate string\n",
-					STRLOC, __PRETTY_FUNCTION__);
+					STRLOC, __func__);
 		abort();
 
 		return NULL;
@@ -233,7 +232,7 @@ LIB_EXPORT char *l_strdup_vprintf(const char *format, va_list args)
 
 	if (len < 0) {
 		fprintf(stderr, "%s:%s(): failed to allocate string\n",
-					STRLOC, __PRETTY_FUNCTION__);
+					STRLOC, __func__);
 		abort();
 
 		return NULL;
@@ -243,256 +242,41 @@ LIB_EXPORT char *l_strdup_vprintf(const char *format, va_list args)
 }
 
 /**
- * l_strfreev:
- * @strlist: String list to free
+ * l_strlcpy:
+ * @dst: Destination buffer for string
+ * @src: Source buffer containing null-terminated string to copy
+ * @len: Maximum destination buffer space to use
  *
- * Frees a list of strings
- **/
-LIB_EXPORT void l_strfreev(char **strlist)
+ * Copies a string from the @src buffer to the @dst buffer, using no
+ * more than @len bytes in @dst. @dst is guaranteed to be
+ * null-terminated. The caller can determine if the copy was truncated by
+ * checking if the return value is greater than or equal to @len.
+ *
+ * NOTE: Passing in a NULL string results in a no-op
+ *
+ * Returns: The length of the @src string, not including the null
+ * terminator.
+ */
+LIB_EXPORT size_t l_strlcpy(char *dst, const char *src, size_t len)
 {
-	l_strv_free(strlist);
-}
+	size_t src_len = src ? strlen(src) : 0;
 
-/**
- * l_strsplit:
- * @str: String to split
- * @sep: The delimiter character
- *
- * Splits a string into pieces which do not contain the delimiter character.
- * As a special case, an empty string is returned as an empty array, e.g.
- * an array with just the NULL element.
- *
- * Note that this function only works with ASCII delimiters.
- *
- * Returns: A newly allocated %NULL terminated string array.  This array
- * should be freed using l_strfreev().
- **/
-LIB_EXPORT char **l_strsplit(const char *str, const char sep)
-{
-	int len;
-	int i;
-	const char *p;
-	char **ret;
+	if (!src)
+		goto done;
 
-	if (unlikely(!str))
-		return NULL;
-
-	if (str[0] == '\0')
-		return l_new(char *, 1);
-
-	for (p = str, len = 1; *p; p++)
-		if (*p == sep)
-			len += 1;
-
-	ret = l_new(char *, len + 1);
-
-	i = 0;
-	p = str;
-	len = 0;
-
-	while (p[len]) {
-		if (p[len] != sep) {
-			len += 1;
-			continue;
+	if (len) {
+		if (src_len < len) {
+			len = src_len + 1;
+		} else {
+			len -= 1;
+			dst[len] = '\0';
 		}
 
-		ret[i++] = l_strndup(p, len);
-		p += len + 1;
-		len = 0;
+		memcpy(dst, src, len);
 	}
 
-	ret[i++] = l_strndup(p, len);
-
-	return ret;
-}
-
-/**
- * l_strsplit_set:
- * @str: String to split
- * @separators: A set of delimiters
- *
- * Splits a string into pieces which do not contain the delimiter characters
- * that can be found in @separators.
- * As a special case, an empty string is returned as an empty array, e.g.
- * an array with just the NULL element.
- *
- * Note that this function only works with ASCII delimiters.
- *
- * Returns: A newly allocated %NULL terminated string array.  This array
- * should be freed using l_strfreev().
- **/
-LIB_EXPORT char **l_strsplit_set(const char *str, const char *separators)
-{
-	int len;
-	int i;
-	const char *p;
-	char **ret;
-	bool sep_table[256];
-
-	if (unlikely(!str))
-		return NULL;
-
-	if (str[0] == '\0')
-		return l_new(char *, 1);
-
-	memset(sep_table, 0, sizeof(sep_table));
-
-	for (p = separators; *p; p++)
-		sep_table[(unsigned char) *p] = true;
-
-	for (p = str, len = 1; *p; p++)
-		if (sep_table[(unsigned char) *p] == true)
-			len += 1;
-
-	ret = l_new(char *, len + 1);
-
-	i = 0;
-	p = str;
-	len = 0;
-
-	while (p[len]) {
-		if (sep_table[(unsigned char) p[len]] != true) {
-			len += 1;
-			continue;
-		}
-
-		ret[i++] = l_strndup(p, len);
-		p += len + 1;
-		len = 0;
-	}
-
-	ret[i++] = l_strndup(p, len);
-
-	return ret;
-}
-
-/**
- * l_strjoinv:
- * @str_array: a %NULL terminated array of strings to join
- * @delim: Delimiting character
- *
- * Joins strings contanied in the @str_array into one long string delimited
- * by @delim.
- *
- * Returns: A newly allocated string that should be freed using l_free()
- */
-LIB_EXPORT char *l_strjoinv(char **str_array, const char delim)
-{
-	size_t len = 0;
-	unsigned int i;
-	char *ret;
-	char *p;
-
-	if (unlikely(!str_array))
-		return NULL;
-
-	if (!str_array[0])
-		return l_strdup("");
-
-	for (i = 0; str_array[i]; i++)
-		len += strlen(str_array[i]);
-
-	len += 1 + i - 1;
-
-	ret = l_malloc(len);
-
-	p = stpcpy(ret, str_array[0]);
-
-	for (i = 1; str_array[i]; i++) {
-		*p++ = delim;
-		p = stpcpy(p, str_array[i]);
-	}
-
-	return ret;
-}
-
-/**
- * l_strv_free:
- * @str_array: a %NULL terminated array of strings
- *
- * Frees strings in @str_array and @str_array itself
- **/
-LIB_EXPORT void l_strv_free(char **str_array)
-{
-	if (likely(str_array)) {
-		int i;
-
-		for (i = 0; str_array[i]; i++)
-			l_free(str_array[i]);
-
-		l_free(str_array);
-	}
-}
-
-/**
- * l_strv_length:
- * @str_array: a %NULL terminated array of strings
- *
- * Returns: the number of strings in @str_array
- */
-LIB_EXPORT unsigned int l_strv_length(char **str_array)
-{
-	unsigned int i = 0;
-
-	if (unlikely(!str_array))
-		return 0;
-
-	while (str_array[i])
-		i += 1;
-
-	return i;
-}
-
-/**
- * l_strv_contains:
- * @str_array: a %NULL terminated array of strings
- * @item: An item to search for, must be not %NULL
- *
- * Returns: #true if @str_array contains item
- */
-LIB_EXPORT bool l_strv_contains(char **str_array, const char *item)
-{
-	unsigned int i = 0;
-
-	if (unlikely(!str_array || !item))
-		return false;
-
-	while (str_array[i]) {
-		if (!strcmp(str_array[i], item))
-			return true;
-
-		i += 1;
-	}
-
-	return false;
-}
-
-/**
- * l_strv_append:
- * @str_array: a %NULL terminated array of strings or %NULL
- * @str: A string to be appened at the end of @str_array
- *
- * Returns: New %NULL terminated array of strings with @str added
- */
-LIB_EXPORT char **l_strv_append(char **str_array, const char *str)
-{
-	char **ret;
-	unsigned int i, len;
-
-	if (unlikely(!str))
-		return str_array;
-
-	len = l_strv_length(str_array);
-	ret = l_new(char *, len + 2);
-
-	for (i = 0; i < len; i++)
-		ret[i] = str_array[i];
-
-	ret[i++] = l_strdup(str);
-
-	l_free(str_array);
-
-	return ret;
+done:
+	return src_len;
 }
 
 /**
@@ -525,38 +309,6 @@ LIB_EXPORT bool l_str_has_prefix(const char *str, const char *prefix)
 	return !strncmp(str, prefix, prefix_len);
 }
 
-/**
- * l_strlcpy:
- * @dst: Destination buffer for string
- * @src: Source buffer containing null-terminated string to copy
- * @len: Maximum destination buffer space to use
- *
- * Copies a string from the @src buffer to the @dst buffer, using no
- * more than @len bytes in @dst. @dst is guaranteed to be
- * null-terminated. The caller can determine if the copy was truncated by
- * checking if the return value is greater than or equal to @len.
- *
- * Returns: The length of the @src string, not including the null
- * terminator.
- */
-LIB_EXPORT size_t l_strlcpy(char* dst, const char *src, size_t len)
-{
-	size_t src_len = strlen(src);
-
-	if (len) {
-		if (src_len < len) {
-			len = src_len + 1;
-		} else {
-			len -= 1;
-			dst[len] = '\0';
-		}
-
-		memcpy(dst, src, len);
-	}
-
-	return src_len;
-}
-
 /**
  * l_str_has_suffix:
  * @str: A string to be examined
@@ -589,16 +341,9 @@ LIB_EXPORT bool l_str_has_suffix(const char *str, const char *suffix)
 	return !strcmp(&str[len_diff], suffix);
 }
 
-/**
- * l_util_hexstring:
- * @buf: buffer pointer
- * @len: length of buffer
- *
- * Returns: a newly allocated hex string
- **/
-LIB_EXPORT char *l_util_hexstring(const unsigned char *buf, size_t len)
+static char *hexstring_common(const unsigned char *buf, size_t len,
+				const char hexdigits[static 16])
 {
-	static const char hexdigits[] = "0123456789abcdef";
 	char *str;
 	size_t i;
 
@@ -617,12 +362,43 @@ LIB_EXPORT char *l_util_hexstring(const unsigned char *buf, size_t len)
 	return str;
 }
 
+/**
+ * l_util_hexstring:
+ * @buf: buffer pointer
+ * @len: length of buffer
+ *
+ * Returns: a newly allocated hex string.  Note that the string will contain
+ * lower case hex digits a-f.  If you require upper case hex digits, use
+ * @l_util_hexstring_upper
+ **/
+LIB_EXPORT char *l_util_hexstring(const unsigned char *buf, size_t len)
+{
+	static const char hexdigits[] = "0123456789abcdef";
+	return hexstring_common(buf, len, hexdigits);
+}
+
+/**
+ * l_util_hexstring_upper:
+ * @buf: buffer pointer
+ * @len: length of buffer
+ *
+ * Returns: a newly allocated hex string.  Note that the string will contain
+ * upper case hex digits a-f.  If you require lower case hex digits, use
+ * @l_util_hexstring
+ **/
+LIB_EXPORT char *l_util_hexstring_upper(const unsigned char *buf, size_t len)
+{
+	static const char hexdigits[] = "0123456789ABCDEF";
+	return hexstring_common(buf, len, hexdigits);
+}
+
 /**
  * l_util_from_hexstring:
  * @str: Null-terminated string containing the hex-encoded bytes
  * @out_len: Number of bytes decoded
  *
- * Returns: a newly allocated byte array
+ * Returns: a newly allocated byte array.  Empty strings are treated as
+ * an error condition.
  **/
 LIB_EXPORT unsigned char *l_util_from_hexstring(const char *str,
 							size_t *out_len)
@@ -636,14 +412,18 @@ LIB_EXPORT unsigned char *l_util_from_hexstring(const char *str,
 		return NULL;
 
 	for (i = 0; str[i]; i++) {
-		c = toupper(str[i]);
+		c = str[i];
 
-		if ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F'))
+		if ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F') ||
+				(c >= 'a' && c <= 'f'))
 			continue;
 
 		return NULL;
 	}
 
+	if (!i)
+		return NULL;
+
 	if ((i % 2) != 0)
 		return NULL;
 
@@ -651,21 +431,24 @@ LIB_EXPORT unsigned char *l_util_from_hexstring(const char *str,
 	buf = l_malloc(i >> 1);
 
 	for (i = 0, j = 0; i < len; i++, j++) {
-		c = toupper(str[i]);
+		c = str[i];
 
 		if (c >= '0' && c <= '9')
 			buf[j] = c - '0';
 		else if (c >= 'A' && c <= 'F')
 			buf[j] = 10 + c - 'A';
+		else if (c >= 'a' && c <= 'f')
+			buf[j] = 10 + c - 'a';
 
 		i += 1;
-
-		c = toupper(str[i]);
+		c = str[i];
 
 		if (c >= '0' && c <= '9')
 			buf[j] = buf[j] * 16 + c - '0';
 		else if (c >= 'A' && c <= 'F')
 			buf[j] = buf[j] * 16 + 10 + c - 'A';
+		else if (c >= 'a' && c <= 'f')
+			buf[j] = buf[j] * 16 + 10 + c - 'a';
 	}
 
 	if (out_len)
@@ -690,7 +473,7 @@ static void hexdump(const char dir, const unsigned char *buf, size_t len,
 		str[((i % 16) * 3) + 1] = ' ';
 		str[((i % 16) * 3) + 2] = hexdigits[buf[i] >> 4];
 		str[((i % 16) * 3) + 3] = hexdigits[buf[i] & 0xf];
-		str[(i % 16) + 51] = isprint(buf[i]) ? buf[i] : '.';
+		str[(i % 16) + 51] = l_ascii_isprint(buf[i]) ? buf[i] : '.';
 
 		if ((i + 1) % 16 == 0) {
 			str[49] = ' ';
@@ -748,6 +531,9 @@ LIB_EXPORT void l_util_hexdumpv(bool in, const struct iovec *iov,
 	size_t c;
 	const uint8_t *buf;
 
+	if (likely(!function))
+		return;
+
 	if (unlikely(!iov || !n_iov))
 		return;
 
@@ -769,7 +555,7 @@ LIB_EXPORT void l_util_hexdumpv(bool in, const struct iovec *iov,
 		str[((i % 16) * 3) + 1] = ' ';
 		str[((i % 16) * 3) + 2] = hexdigits[buf[c] >> 4];
 		str[((i % 16) * 3) + 3] = hexdigits[buf[c] & 0xf];
-		str[(i % 16) + 51] = isprint(buf[c]) ? buf[c] : '.';
+		str[(i % 16) + 51] = l_ascii_isprint(buf[c]) ? buf[c] : '.';
 
 		if ((i + 1) % 16 == 0) {
 			str[49] = ' ';
@@ -854,3 +640,27 @@ LIB_EXPORT const char *l_util_get_debugfs_path(void)
 
 	return path;
 }
+
+LIB_EXPORT bool l_memeq(const void *field, size_t size, uint8_t byte)
+{
+	const uint8_t *mem = field;
+	size_t i;
+
+	for (i = 0; i < size; i++)
+		if (mem[i] != byte)
+			return false;
+
+	return true;
+}
+
+LIB_EXPORT bool l_secure_memeq(const void *field, size_t size, uint8_t byte)
+{
+	const volatile uint8_t *mem = field;
+	size_t i;
+	bool diff = false;
+
+	for (i = 0; i < size; i++)
+		diff |= mem[i] != byte;
+
+	return !diff;
+}
