@@ -35,6 +35,7 @@
 #include <ell/ell.h>
 
 #include "src/iwd.h"
+#include "src/module.h"
 #include "src/storage.h"
 #include "src/common.h"
 #include "src/network.h"
@@ -334,27 +335,31 @@ bool network_info_match_nai_realm(const struct network_info *info,
 	return info->ops->match_nai_realms(info, nai_realms);
 }
 
-void known_network_update(struct network_info *network,
-					struct l_settings *settings,
+void known_network_set_connected_time(struct network_info *network,
 					uint64_t connected_time)
 {
-	bool is_hidden = false;
-	bool is_autoconnectable;
-
-	if (network->connected_time != connected_time) {
-		l_dbus_property_changed(dbus_get_bus(),
-					known_network_get_path(network),
-					IWD_KNOWN_NETWORK_INTERFACE,
-					"LastConnectedTime");
-
-		l_queue_remove(known_networks, network);
-		l_queue_insert(known_networks, network, connected_time_compare,
-				NULL);
-	}
+	if (network->connected_time == connected_time)
+		return;
 
 	network->connected_time = connected_time;
 
-	l_settings_get_bool(settings, "Settings", "Hidden", &is_hidden);
+	l_dbus_property_changed(dbus_get_bus(),
+				known_network_get_path(network),
+				IWD_KNOWN_NETWORK_INTERFACE,
+				"LastConnectedTime");
+
+	l_queue_remove(known_networks, network);
+	l_queue_insert(known_networks, network, connected_time_compare, NULL);
+}
+
+void known_network_update(struct network_info *network,
+					struct l_settings *settings)
+{
+	bool is_hidden;
+	bool is_autoconnectable;
+
+	if (!l_settings_get_bool(settings, "Settings", "Hidden", &is_hidden))
+		is_hidden = false;
 
 	if (network->is_hidden != is_hidden) {
 		if (network->is_hidden && !is_hidden)
@@ -751,10 +756,11 @@ static void known_networks_watch_cb(const char *filename,
 		if (settings) {
 			connected_time = l_path_get_mtime(full_path);
 
-			if (network_before)
-				known_network_update(network_before, settings,
+			if (network_before) {
+				known_network_set_connected_time(network_before,
 							connected_time);
-			else
+				known_network_update(network_before, settings);
+			} else
 				known_network_new(ssid, security, settings,
 							connected_time);
 		} else if (network_before)
@@ -764,6 +770,14 @@ static void known_networks_watch_cb(const char *filename,
 
 		break;
 	case L_DIR_WATCH_EVENT_ACCESSED:
+		break;
+	case L_DIR_WATCH_EVENT_ATTRIB:
+		if (network_before) {
+			connected_time = l_path_get_mtime(full_path);
+			known_network_set_connected_time(network_before,
+								connected_time);
+		}
+
 		break;
 	}
 }
@@ -896,29 +910,38 @@ static int known_network_frequencies_load(void)
 		const char *path = l_settings_get_value(known_freqs, groups[i],
 							"name");
 		if (!path)
-			continue;
+			goto invalid_entry;
 
 		info = find_network_info_from_path(path);
 		if (!info)
-			continue;
+			goto invalid_entry;
+
+		if (info->has_uuid)
+			goto invalid_entry;
 
 		freq_list = l_settings_get_string(known_freqs, groups[i],
 							"list");
 		if (!freq_list)
-			continue;
+			goto invalid_entry;
 
 		known_frequencies = known_frequencies_from_string(freq_list);
-		if (!known_frequencies)
-			goto next;
+		l_free(freq_list);
 
-		if (!l_uuid_from_string(groups[i], uuid))
-			goto next;
+		if (!known_frequencies)
+			goto invalid_entry;
+
+		if (!l_uuid_from_string(groups[i], uuid)) {
+			l_queue_destroy(known_frequencies, l_free);
+			goto invalid_entry;
+		}
 
 		network_info_set_uuid(info, uuid);
 		info->known_frequencies = known_frequencies;
 
-next:
-		l_free(freq_list);
+		continue;
+
+invalid_entry:
+		l_settings_remove_group(known_freqs, groups[i]);
 	}
 
 	l_strv_free(groups);
