@@ -31,6 +31,8 @@
 
 #include <ell/ell.h>
 
+#include "src/util.h"
+#include "src/ie.h"
 #include "src/wscutil.h"
 
 const unsigned char wsc_wfa_oui[3] = { 0x00, 0x37, 0x2a };
@@ -360,13 +362,25 @@ static bool extract_network_key(struct wsc_attr_iter *iter, void *data)
 {
 	struct iovec *network_key = data;
 	unsigned int len;
+	const uint8_t *key;
 
 	len = wsc_attr_iter_get_length(iter);
 	if (len > 64)
 		return false;
 
+	/*
+	 * WSC 2.0.5, Section 12, Network Key:
+	 * "Some existing implementations based on v1.0h null-terminate the
+	 * passphrase value, i.e., add an extra 0x00 octet into the end of
+	 * the value. For backwards compatibility, implementations shall be
+	 * able to parse such a value"
+	 */
+	key = wsc_attr_iter_get_data(iter);
+	if (len && key[len - 1] == 0x00)
+		len--;
+
 	network_key->iov_len = len;
-	network_key->iov_base = (void *) wsc_attr_iter_get_data(iter);
+	network_key->iov_base = (void *) key;
 
 	return true;
 }
@@ -1620,11 +1634,12 @@ static bool wsc_attr_builder_start_attr(struct wsc_attr_builder *builder,
 	/* Record previous attribute's length */
 	if (builder->curlen > 0) {
 		bytes = builder->buf + builder->offset;
-		l_put_be16(builder->curlen, bytes + 2);
-		builder->offset += 4 + builder->curlen;
-		builder->curlen = 0;
+		l_put_be16(builder->curlen - 4, bytes + 2);
+		builder->offset += builder->curlen;
 	}
 
+	builder->curlen = 4;
+
 	if (builder->offset + 4 >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
@@ -1636,10 +1651,10 @@ static bool wsc_attr_builder_start_attr(struct wsc_attr_builder *builder,
 
 static bool wsc_attr_builder_put_u8(struct wsc_attr_builder *builder, uint8_t v)
 {
-	if (builder->offset + 4 + builder->curlen + 1 >= builder->capacity)
+	if (builder->offset + builder->curlen + 1 >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
-	builder->buf[builder->offset + 4 + builder->curlen] = v;
+	builder->buf[builder->offset + builder->curlen] = v;
 	builder->curlen += 1;
 
 	return true;
@@ -1648,10 +1663,10 @@ static bool wsc_attr_builder_put_u8(struct wsc_attr_builder *builder, uint8_t v)
 static bool wsc_attr_builder_put_u16(struct wsc_attr_builder *builder,
 								uint16_t v)
 {
-	if (builder->offset + 4 + builder->curlen + 2 >= builder->capacity)
+	if (builder->offset + builder->curlen + 2 >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
-	l_put_be16(v, builder->buf + builder->offset + 4 + builder->curlen);
+	l_put_be16(v, builder->buf + builder->offset + builder->curlen);
 	builder->curlen += 2;
 
 	return true;
@@ -1660,10 +1675,10 @@ static bool wsc_attr_builder_put_u16(struct wsc_attr_builder *builder,
 static bool wsc_attr_builder_put_u32(struct wsc_attr_builder *builder,
 								uint32_t v)
 {
-	if (builder->offset + 4 + builder->curlen + 4 >= builder->capacity)
+	if (builder->offset + builder->curlen + 4 >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
-	l_put_be32(v, builder->buf + builder->offset + 4 + builder->curlen);
+	l_put_be32(v, builder->buf + builder->offset + builder->curlen);
 	builder->curlen += 4;
 
 	return true;
@@ -1672,12 +1687,10 @@ static bool wsc_attr_builder_put_u32(struct wsc_attr_builder *builder,
 static bool wsc_attr_builder_put_bytes(struct wsc_attr_builder *builder,
 					const void *bytes, size_t size)
 {
-	while (builder->offset + 4 + builder->curlen + size >=
-							builder->capacity)
+	while (builder->offset + builder->curlen + size >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
-	memcpy(builder->buf + builder->offset + 4 + builder->curlen,
-								bytes, size);
+	memcpy(builder->buf + builder->offset + builder->curlen, bytes, size);
 	builder->curlen += size;
 
 	return true;
@@ -1686,10 +1699,10 @@ static bool wsc_attr_builder_put_bytes(struct wsc_attr_builder *builder,
 static bool wsc_attr_builder_put_oui(struct wsc_attr_builder *builder,
 							const uint8_t *oui)
 {
-	if (builder->offset + 4 + builder->curlen + 3 >= builder->capacity)
+	if (builder->offset + builder->curlen + 3 >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
-	memcpy(builder->buf + builder->offset + 4 + builder->curlen, oui, 3);
+	memcpy(builder->buf + builder->offset + builder->curlen, oui, 3);
 	builder->curlen += 3;
 
 	return true;
@@ -1707,11 +1720,10 @@ static bool wsc_attr_builder_put_string(struct wsc_attr_builder *builder,
 		len = 1;
 	}
 
-	if (builder->offset + 4 + builder->curlen + len >= builder->capacity)
+	if (builder->offset + builder->curlen + len >= builder->capacity)
 		wsc_attr_builder_grow(builder);
 
-	memcpy(builder->buf + builder->offset + 4 + builder->curlen,
-								string, len);
+	memcpy(builder->buf + builder->offset + builder->curlen, string, len);
 	builder->curlen += len;
 
 	return true;
@@ -1739,8 +1751,8 @@ static uint8_t *wsc_attr_builder_free(struct wsc_attr_builder *builder,
 
 	if (builder->curlen > 0) {
 		uint8_t *bytes = builder->buf + builder->offset;
-		l_put_be16(builder->curlen, bytes + 2);
-		builder->offset += 4 + builder->curlen;
+		l_put_be16(builder->curlen - 4, bytes + 2);
+		builder->offset += builder->curlen;
 		builder->curlen = 0;
 	}
 
@@ -1759,6 +1771,12 @@ static uint8_t *wsc_attr_builder_free(struct wsc_attr_builder *builder,
 	return ret;
 }
 
+static void build_ap_setup_locked(struct wsc_attr_builder *builder, bool locked)
+{
+	wsc_attr_builder_start_attr(builder, WSC_ATTR_AP_SETUP_LOCKED);
+	wsc_attr_builder_put_u8(builder, locked ? 0x01 : 0x00);
+}
+
 static void build_association_state(struct wsc_attr_builder *builder,
 					enum wsc_association_state state)
 {
@@ -2014,6 +2032,22 @@ static void build_r_snonce2(struct wsc_attr_builder *builder,
 	wsc_attr_builder_put_bytes(builder, nonce, 16);
 }
 
+static void build_selected_registrar(struct wsc_attr_builder *builder,
+							bool selected)
+{
+	wsc_attr_builder_start_attr(builder, WSC_ATTR_SELECTED_REGISTRAR);
+	wsc_attr_builder_put_u8(builder, selected ? 0x01 : 0x00);
+}
+
+static void build_selected_registrar_configuration_methods(
+					struct wsc_attr_builder *builder,
+					uint16_t config_methods)
+{
+	wsc_attr_builder_start_attr(builder,
+			WSC_ATTR_SELECTED_REGISTRAR_CONFIGURATION_METHODS);
+	wsc_attr_builder_put_u16(builder, config_methods);
+}
+
 static void build_ssid(struct wsc_attr_builder *builder, const uint8_t *ssid,
 							size_t ssid_len)
 {
@@ -2060,6 +2094,20 @@ static void build_wsc_state(struct wsc_attr_builder *builder,
 	wsc_attr_builder_put_u8(builder, 1);				\
 	wsc_attr_builder_put_u8(builder, 0x20)
 
+static void wfa_build_authorized_macs(struct wsc_attr_builder *builder,
+				const uint8_t authorized_macs[static 30])
+{
+	int count;
+
+	for (count = 1; count < 5; count++)
+		if (util_mem_is_zero(authorized_macs + count * 6, 6))
+			break;
+
+	wsc_attr_builder_put_u8(builder, WSC_WFA_EXTENSION_AUTHORIZED_MACS);
+	wsc_attr_builder_put_u8(builder, count * 6);
+	wsc_attr_builder_put_bytes(builder, authorized_macs, count * 6);
+}
+
 uint8_t *wsc_build_credential(const struct wsc_credential *in, size_t *out_len)
 {
 	struct wsc_attr_builder *builder;
@@ -2090,6 +2138,53 @@ static void build_credential(struct wsc_attr_builder *builder,
 	l_free(data);
 }
 
+uint8_t *wsc_build_beacon(const struct wsc_beacon *beacon, size_t *out_len)
+{
+	struct wsc_attr_builder *builder;
+	uint8_t *ret;
+
+	builder = wsc_attr_builder_new(512);
+	build_version(builder, 0x10);
+	build_wsc_state(builder, beacon->state);
+
+	if (beacon->ap_setup_locked)
+		build_ap_setup_locked(builder, true);
+
+	if (beacon->selected_registrar) {
+		build_selected_registrar(builder, true);
+		build_device_password_id(builder, beacon->device_password_id);
+		build_selected_registrar_configuration_methods(builder,
+					beacon->selected_reg_config_methods);
+	}
+
+	/* These two "should be provided" if dual-band */
+	if (__builtin_popcount(beacon->rf_bands) > 1) {
+		if (beacon->selected_registrar)
+			build_uuid_e(builder, beacon->uuid_e);
+
+		build_rf_bands(builder, beacon->rf_bands);
+	}
+
+	if (!beacon->version2)
+		goto done;
+
+	START_WFA_VENDOR_EXTENSION();
+
+	if (!util_mem_is_zero(beacon->authorized_macs, 6))
+		wfa_build_authorized_macs(builder, beacon->authorized_macs);
+
+	if (beacon->reg_config_methods) {
+		wsc_attr_builder_put_u8(builder,
+			WSC_WFA_EXTENSION_REGISTRAR_CONFIGRATION_METHODS);
+		wsc_attr_builder_put_u8(builder, 2);
+		wsc_attr_builder_put_u16(builder, beacon->reg_config_methods);
+	}
+
+done:
+	ret = wsc_attr_builder_free(builder, false, out_len);
+	return ret;
+}
+
 uint8_t *wsc_build_probe_request(const struct wsc_probe_request *probe_request,
 							size_t *out_len)
 {
@@ -2129,6 +2224,64 @@ done:
 	return ret;
 }
 
+uint8_t *wsc_build_probe_response(
+		const struct wsc_probe_response *probe_response,
+		size_t *out_len)
+{
+	struct wsc_attr_builder *builder;
+	uint8_t *ret;
+
+	builder = wsc_attr_builder_new(512);
+	build_version(builder, 0x10);
+	build_wsc_state(builder, probe_response->state);
+
+	if (probe_response->ap_setup_locked)
+		build_ap_setup_locked(builder, true);
+
+	if (probe_response->selected_registrar) {
+		build_selected_registrar(builder, true);
+		build_device_password_id(builder,
+				probe_response->device_password_id);
+		build_selected_registrar_configuration_methods(builder,
+				probe_response->selected_reg_config_methods);
+	}
+
+	build_response_type(builder, probe_response->response_type);
+	build_uuid_e(builder, probe_response->uuid_e);
+	build_manufacturer(builder, probe_response->manufacturer);
+	build_model_name(builder, probe_response->model_name);
+	build_model_number(builder, probe_response->model_number);
+	build_serial_number(builder, probe_response->serial_number);
+	build_primary_device_type(builder,
+					&probe_response->primary_device_type);
+	build_device_name(builder, probe_response->device_name);
+	build_configuration_methods(builder, probe_response->config_methods);
+
+	if (__builtin_popcount(probe_response->rf_bands) > 1)
+		build_rf_bands(builder, probe_response->rf_bands);
+
+	if (!probe_response->version2)
+		goto done;
+
+	START_WFA_VENDOR_EXTENSION();
+
+	if (!util_mem_is_zero(probe_response->authorized_macs, 6))
+		wfa_build_authorized_macs(builder,
+					probe_response->authorized_macs);
+
+	if (probe_response->reg_config_methods) {
+		wsc_attr_builder_put_u8(builder,
+			WSC_WFA_EXTENSION_REGISTRAR_CONFIGRATION_METHODS);
+		wsc_attr_builder_put_u8(builder, 2);
+		wsc_attr_builder_put_u16(builder,
+					probe_response->reg_config_methods);
+	}
+
+done:
+	ret = wsc_attr_builder_free(builder, false, out_len);
+	return ret;
+}
+
 uint8_t *wsc_build_association_request(
 		const struct wsc_association_request *association_request,
 		size_t *out_len)
@@ -2698,3 +2851,202 @@ bool wsc_pin_generate(char *pin)
 
 	return true;
 }
+
+struct device_type_category_info {
+	const char *category_str;
+	unsigned int subcategory_max;
+	const char **subcategory_str;
+};
+
+/* WSC 2.0.5, Table 41 strings adapted to IWD DBus enum convention */
+struct device_type_category_info device_type_categories[] = {
+	[1] = {
+		"computer",
+		10,
+		(const char *[]) {
+			[1] = "pc",
+			[2] = "server",
+			[3] = "media-center",
+			[4] = "ultra-mobile-pc",
+			[5] = "notebook",
+			[6] = "desktop",
+			[7] = "mobile-internet-device",
+			[8] = "netbook",
+			[9] = "tablet",
+			[10] = "ultrabook",
+		},
+	},
+	[2] = {
+		"input-device",
+		9,
+		(const char *[]) {
+			[1] = "keyboard",
+			[2] = "mouse",
+			[3] = "joystick",
+			[4] = "trackball",
+			[5] = "gaming-controller",
+			[6] = "remote",
+			[7] = "touchscreen",
+			[8] = "biometric-reader",
+			[9] = "barcode-reader",
+		},
+	},
+	[3] = {
+		"printer-scanner",
+		5,
+		(const char *[]) {
+			[1] = "printer-print-server",
+			[2] = "scanner",
+			[3] = "fax",
+			[4] = "copier",
+			[5] = "printer-scanner-fax-copier",
+		},
+	},
+	[4] = {
+		"camera",
+		4,
+		(const char *[]) {
+			[1] = "digital-still-camera",
+			[2] = "video-camera",
+			[3] = "web-camera",
+			[4] = "security-camera",
+		},
+	},
+	[5] = {
+		"storage",
+		1,
+		(const char *[]) {
+			[1] = "nas",
+		},
+	},
+	[6] = {
+		"network-infrastructure",
+		5,
+		(const char *[]) {
+			[1] = "ap",
+			[2] = "router",
+			[3] = "switch",
+			[4] = "gateway",
+			[5] = "bridge",
+		},
+	},
+	[7] = {
+		"display",
+		4,
+		(const char *[]) {
+			[1] = "television",
+			[2] = "electronic-picture-frame",
+			[3] = "projector",
+			[4] = "monitor",
+		},
+	},
+	[8] = {
+		"multimedia-device",
+		6,
+		(const char *[]) {
+			[1] = "dar",
+			[2] = "pvr",
+			[3] = "mcx",
+			[4] = "set-top-box",
+			[5] = "media-server-adapter-extender",
+			[6] = "portable-video-player",
+		},
+	},
+	[9] = {
+		"gaming-device",
+		5,
+		(const char *[]) {
+			[1] = "xbox",
+			[2] = "xbox360",
+			[3] = "playstation",
+			[4] = "game-console-adapter",
+			[5] = "portable-gaming-device",
+		},
+	},
+	[10] = {
+		"telephone",
+		5,
+		(const char *[]) {
+			[1] = "windows-mobile",
+			[2] = "single-mode-phone",
+			[3] = "dual-mode-phone",
+			[4] = "single-mode-smartphone",
+			[5] = "dual-mode-smartphone",
+		},
+	},
+	[11] = {
+		"audio-device",
+		7,
+		(const char *[]) {
+			[1] = "audio-tuner-receiver",
+			[2] = "speakers",
+			[3] = "portable-music-player",
+			[4] = "headset",
+			[5] = "headphones",
+			[6] = "microphone",
+			[7] = "home-theater-system",
+		},
+	},
+	[12] = {
+		"docking-device",
+		2,
+		(const char *[]) {
+			[1] = "computer-docking-station",
+			[2] = "media-kiosk",
+		},
+	},
+};
+
+bool wsc_device_type_to_dbus_str(const struct wsc_primary_device_type *val,
+					const char **category_str,
+					const char **subcategory_str)
+{
+	struct device_type_category_info *cat;
+
+	if (val->category >= L_ARRAY_SIZE(device_type_categories))
+		return false;
+
+	cat = &device_type_categories[val->category];
+
+	if (!cat->category_str)
+		return false;
+
+	if (category_str)
+		*category_str = cat->category_str;
+
+	if (!subcategory_str)
+		return true;
+
+	if (memcmp(val->oui, microsoft_oui, 3) || val->oui_type != 4)
+		*subcategory_str = NULL;	/* Vendor-specific */
+	else if (val->subcategory <= cat->subcategory_max &&
+			cat->subcategory_str[val->subcategory])
+		*subcategory_str = cat->subcategory_str[val->subcategory];
+	else
+		*subcategory_str = NULL;	/* Unknown */
+
+	return true;
+}
+
+bool wsc_device_type_from_subcategory_str(struct wsc_primary_device_type *out,
+						const char *subcategory_str)
+{
+	struct device_type_category_info *cat = device_type_categories + 1;
+	unsigned int i;
+
+	for (i = 1; i < L_ARRAY_SIZE(device_type_categories); i++, cat++) {
+		unsigned int j;
+
+		for (j = 1; j <= cat->subcategory_max; j++)
+			if (!strcasecmp(subcategory_str,
+					cat->subcategory_str[j])) {
+				out->category = i;
+				memcpy(out->oui, microsoft_oui, 3);
+				out->oui_type = 4;
+				out->subcategory = j;
+				return true;
+			}
+	}
+
+	return false;
+}
